theory Syntax
imports Environ

begin

datatype qual = qL | qU

datatype pty =
    TyUnit
  | TyBase name
  | TyArrow ty ty    (infixr "\<rightarrow>" 900)
  | TyRef ty

and ty = Qual qual pty

datatype tm =
    TmUnit qual
  | TmFreeVar name
  | TmVar nat
  | TmAbs qual ty tm  ("\<lambda>^_[_]._" [900,900,800] 800)
  | TmApp tm tm       (infixl "\<cdot>" 900)
  | TmNew tm
  | TmFree tm
  | TmAssign tm tm    (infixr ":=" 900)
  | TmSwap tm tm      (infixr ":=:" 900)
  | TmDeref tm
  | TmLoc qual name

(* vsubst: freshening substitution *)
consts vsubst_tm:: "tm \<Rightarrow> nat \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "vsubst_tm s i (TmUnit q) = TmUnit q"
  "vsubst_tm s i (TmFreeVar b) = TmFreeVar b"
  "vsubst_tm s i (TmVar n) = (if i = n then s else (TmVar n))"
  "vsubst_tm s i (\<lambda>^q[\<tau>].e) = \<lambda>^q[\<tau>].(vsubst_tm s (Suc i) e)"
  "vsubst_tm s i (e1\<cdot>e2) = (vsubst_tm s i e1)\<cdot>(vsubst_tm s i e2)"
  "vsubst_tm s i (TmNew e) = TmNew (vsubst_tm s i e)"
  "vsubst_tm s i (TmFree e) = TmFree (vsubst_tm s i e)"
  "vsubst_tm s i (TmAssign e1 e2) = TmAssign (vsubst_tm s i e1) (vsubst_tm s i e2)"
  "vsubst_tm s i (TmSwap e1 e2) = TmSwap (vsubst_tm s i e1) (vsubst_tm s i e2)"
  "vsubst_tm s i (TmDeref e) = TmDeref (vsubst_tm s i e)"
  "vsubst_tm s i (TmLoc q l) = TmLoc q l"

consts freshen_tm :: "name \<Rightarrow> tm \<Rightarrow> tm"
defs freshen_tm[simp]: "freshen_tm x e \<equiv> vsubst_tm (TmFreeVar x) 0 e"

consts free_tm :: "tm \<Rightarrow> name list"
primrec
  "free_tm (TmUnit q)            = []"
  "free_tm (TmFreeVar b)         = [b]"
  "free_tm (TmVar n)             = []"
  "free_tm (\<lambda>^q[\<tau>].e)            = free_tm e"
  "free_tm (e1\<cdot>e2)               = (free_tm e1)@(free_tm e2)"
  "free_tm (TmNew e)             = free_tm e"
  "free_tm (TmFree e)            = free_tm e"
  "free_tm (TmAssign e1 e2)      = (free_tm e1)@(free_tm e2)"
  "free_tm (TmSwap e1 e2)        = (free_tm e1)@(free_tm e2)"
  "free_tm (TmDeref e)           = free_tm e"
  "free_tm (TmLoc q l)           = []"

syntax
  is_free_in_tm :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free in _" [900,900] 91)
translations
  "x free in e" \<rightleftharpoons> "x \<in> set (free_tm e)"

consts free_loc :: "tm \<Rightarrow> name list"
primrec
  "free_loc (TmUnit q)            = []"
  "free_loc (TmFreeVar b)         = []"
  "free_loc (TmVar n)             = []"
  "free_loc (\<lambda>^q[\<tau>].e)            = free_loc e"
  "free_loc (e1\<cdot>e2)               = (free_loc e1)@(free_loc e2)"
  "free_loc (TmNew e)             = free_loc e"
  "free_loc (TmFree e)            = free_loc e"
  "free_loc (TmAssign e1 e2)      = (free_loc e1)@(free_loc e2)"
  "free_loc (TmSwap e1 e2)        = (free_loc e1)@(free_loc e2)"
  "free_loc (TmDeref e)           = free_loc e"
  "free_loc (TmLoc q l)           = [l]"

syntax
  is_free_in_loc :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free loc in _" [900,900] 91)
translations
  "x free loc in e" \<rightleftharpoons> "x \<in> set (free_loc e)"

consts shift_many_db_list :: "nat \<Rightarrow> nat list \<Rightarrow> nat list"
primrec
  "shift_many_db_list k [] = []"
  "shift_many_db_list k (n#l) = (if n < k then shift_many_db_list k l
                                          else (n - k)#(shift_many_db_list k l))"

consts shift_db_list :: "nat list \<Rightarrow> nat list"
defs shift_db_list[simp]: "shift_db_list l \<equiv> shift_many_db_list 1 l"

consts free_db_tm :: "tm \<Rightarrow> nat list"
primrec
  "free_db_tm (TmUnit q)       = []"
  "free_db_tm (TmFreeVar b)    = []"
  "free_db_tm (TmVar n)        = [n]"
  "free_db_tm (\<lambda>^q[\<tau>]. e)      = shift_db_list (free_db_tm e)"
  "free_db_tm (e1\<cdot>e2)          = (free_db_tm e1)@(free_db_tm e2)"
  "free_db_tm (TmNew e)        = free_db_tm e"
  "free_db_tm (TmFree e)       = free_db_tm e"
  "free_db_tm (TmAssign e1 e2) = (free_db_tm e1)@(free_db_tm e2)"
  "free_db_tm (TmSwap e1 e2)   = (free_db_tm e1)@(free_db_tm e2)"
  "free_db_tm (TmDeref e)      = free_db_tm e"
  "free_db_tm (TmLoc q l)      = []"

lemma not_free_subst:
  assumes "\<not> x free in e"
  and "\<not> x free in e'"
  shows "\<not> x free in (vsubst_tm e' k e)"
using prems by (induct e arbitrary: k, auto)

lemma free_subst:
  assumes "x free in e"
  shows "x free in (vsubst_tm e' k e)"
using prems by (induct e arbitrary: k, auto)

lemma vsubst_var_same_size[simp]:
  shows "size (vsubst_tm (TmFreeVar x) k e) = size e"
by (induct e arbitrary: k, auto)

lemma subst_subst_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_tm (TmFreeVar x) k e_s = e_s"
  shows "vsubst_tm (TmFreeVar x) i (vsubst_tm e_s j e) =
         vsubst_tm e_s j (vsubst_tm (TmFreeVar x) i e)"
using prems by (induct e arbitrary: i j, auto)

lemma vsubst_tm_same_loc:
  shows "free_loc e = free_loc (vsubst_tm (TmFreeVar x) i e)"
by (induct e arbitrary: i, auto)

lemma not_free_subst_same_subst:
  assumes "\<not> x free in vsubst_tm (TmFreeVar x) i e"
  shows "vsubst_tm e_s i e = e"
using prems by (induct e arbitrary: i, auto)

end
